@@ -6,7 +6,7 @@ use std::{
 use crate::{
     date::{Date, InfiniteOrNanDate},
     error::{Error, ErrorKind},
-    stream::Event,
+    stream::{Event, OwnedEvent},
     u64_to_usize, Uid,
 };
 
@@ -187,7 +187,7 @@ impl<R: Read + Seek> BinaryReader<R> {
         if offset >= self.trailer_start_offset {
             return Err(self.with_pos(ErrorKind::ObjectOffsetTooLarge));
         }
-        Ok(self.reader.seek(SeekFrom::Start(offset))?)
+        self.reader.seek(SeekFrom::Start(offset))
     }
 
     fn push_stack_item_and_check_for_recursion(&mut self, item: StackItem) -> Result<(), Error> {
@@ -208,7 +208,7 @@ impl<R: Read + Seek> BinaryReader<R> {
         item
     }
 
-    fn read_next(&mut self) -> Result<Option<Event>, Error> {
+    fn read_next(&mut self) -> Result<Option<OwnedEvent>, Error> {
         let object_ref = if self.ref_size == 0 {
             // Initialise here rather than in new
             self.read_trailer()?;
@@ -268,7 +268,7 @@ impl<R: Read + Seek> BinaryReader<R> {
             (0x4, n) => {
                 // Data
                 let len = self.read_object_len(n)?;
-                Some(Event::Data(self.read_data(len)?))
+                Some(Event::Data(self.read_data(len)?.into()))
             }
             (0x5, n) => {
                 // ASCII string
@@ -276,7 +276,7 @@ impl<R: Read + Seek> BinaryReader<R> {
                 let raw = self.read_data(len)?;
                 let string = String::from_utf8(raw)
                     .map_err(|_| self.with_pos(ErrorKind::InvalidUtf8String))?;
-                Some(Event::String(string))
+                Some(Event::String(string.into()))
             }
             (0x6, n) => {
                 // UTF-16 string
@@ -289,7 +289,7 @@ impl<R: Read + Seek> BinaryReader<R> {
 
                 let string = String::from_utf16(&raw_utf16)
                     .map_err(|_| self.with_pos(ErrorKind::InvalidUtf16String))?;
-                Some(Event::String(string))
+                Some(Event::String(string.into()))
             }
             (0x8, n) if n < 8 => {
                 // Uid
@@ -392,9 +392,9 @@ impl<R: Read + Seek> BinaryReader<R> {
 }
 
 impl<R: Read + Seek> Iterator for BinaryReader<R> {
-    type Item = Result<Event, Error>;
+    type Item = Result<OwnedEvent, Error>;
 
-    fn next(&mut self) -> Option<Result<Event, Error>> {
+    fn next(&mut self) -> Option<Result<OwnedEvent, Error>> {
         match self.read_next() {
             Ok(Some(event)) => Some(Ok(event)),
             Err(err) => {
@@ -409,7 +409,7 @@ impl<R: Read + Seek> Iterator for BinaryReader<R> {
 
 #[cfg(test)]
 mod tests {
-    use std::{fs::File, path::Path};
+    use std::fs::File;
 
     use super::*;
     use crate::{stream::Event, Uid};
@@ -418,7 +418,7 @@ mod tests {
     fn streaming_parser() {
         use crate::stream::Event::*;
 
-        let reader = File::open(&Path::new("./tests/data/binary.plist")).unwrap();
+        let reader = File::open("./tests/data/binary.plist").unwrap();
         let streaming_parser = BinaryReader::new(reader);
         let events: Vec<Event> = streaming_parser.map(|e| e.unwrap()).collect();
 
@@ -427,7 +427,7 @@ mod tests {
             String("Author".into()),
             String("William Shakespeare".into()),
             String("Birthdate".into()),
-            Date(super::Date::from_rfc3339("1981-05-16T11:32:06Z").unwrap()),
+            Date(super::Date::from_xml_format("1981-05-16T11:32:06Z").unwrap()),
             String("EmptyArray".into()),
             StartArray(Some(0)),
             EndCollection,
@@ -442,7 +442,7 @@ mod tests {
             Real(1.6),
             String("Lines".into()),
             StartArray(Some(2)),
-            String("It is a tale told by an idiot,".into()),
+            String("It is a tale told by an idiot,     ".into()),
             String("Full of sound and fury, signifying nothing.".into()),
             EndCollection,
             String("Death".into()),
@@ -454,7 +454,7 @@ mod tests {
             String("IsTrue".into()),
             Boolean(true),
             String("Data".into()),
-            Data(vec![0, 0, 0, 190, 0, 0, 0, 3, 0, 0, 0, 30, 0, 0, 0]),
+            Data(vec![0, 0, 0, 190, 0, 0, 0, 3, 0, 0, 0, 30, 0, 0, 0].into()),
             EndCollection,
         ];
 
@@ -463,11 +463,11 @@ mod tests {
 
     #[test]
     fn utf16_plist() {
-        let reader = File::open(&Path::new("./tests/data/utf16_bplist.plist")).unwrap();
+        let reader = File::open("./tests/data/utf16_bplist.plist").unwrap();
         let streaming_parser = BinaryReader::new(reader);
         let mut events: Vec<Event> = streaming_parser.map(|e| e.unwrap()).collect();
 
-        assert_eq!(events[2], Event::String("\u{2605} or better".to_owned()));
+        assert_eq!(events[2], Event::String("\u{2605} or better".into()));
 
         let poem = if let Event::String(ref mut poem) = events[4] {
             poem
@@ -475,12 +475,12 @@ mod tests {
             panic!("not a string")
         };
         assert_eq!(poem.len(), 643);
-        assert_eq!(poem.pop().unwrap(), '\u{2605}');
+        assert_eq!(poem.to_mut().pop().unwrap(), '\u{2605}');
     }
 
     #[test]
     fn nskeyedarchiver_plist() {
-        let reader = File::open(&Path::new("./tests/data/binary_NSKeyedArchiver.plist")).unwrap();
+        let reader = File::open("./tests/data/binary_NSKeyedArchiver.plist").unwrap();
         let streaming_parser = BinaryReader::new(reader);
         let events: Vec<Event> = streaming_parser.map(|e| e.unwrap()).collect();
 
