@@ -1,19 +1,17 @@
-use base64;
-use line_wrap;
-use std::{borrow::Cow, io::Write};
-use xml_rs::{
-    name::Name,
-    namespace::Namespace,
-    writer::{EmitterConfig, Error as XmlWriterError, EventWriter, XmlEvent},
+use base64::{engine::general_purpose::STANDARD as base64_standard, Engine};
+use quick_xml::{
+    events::{BytesEnd, BytesStart, BytesText, Event as XmlEvent},
+    Error as XmlWriterError, Writer as EventWriter,
 };
+use std::{borrow::Cow, io::Write};
 
 use crate::{
     error::{self, Error, ErrorKind, EventKind},
-    stream::Writer,
+    stream::{Writer, XmlWriteOptions},
     Date, Integer, Uid,
 };
 
-static XML_PROLOGUE: &str = r#"<?xml version="1.0" encoding="UTF-8"?>
+static XML_PROLOGUE: &[u8] = br#"<?xml version="1.0" encoding="UTF-8"?>
 <!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
 <plist version="1.0">
 "#;
@@ -26,34 +24,47 @@ enum Element {
 
 pub struct XmlWriter<W: Write> {
     xml_writer: EventWriter<W>,
+    write_root_element: bool,
+    started_plist: bool,
     stack: Vec<Element>,
     expecting_key: bool,
-    written_prologue: bool,
-    // Not very nice
-    empty_namespace: Namespace,
+    pending_collection: Option<PendingCollection>,
+}
+
+enum PendingCollection {
+    Array,
+    Dictionary,
 }
 
 impl<W: Write> XmlWriter<W> {
+    #[cfg(feature = "enable_unstable_features_that_may_break_with_minor_version_bumps")]
     pub fn new(writer: W) -> XmlWriter<W> {
-        let config = EmitterConfig::new()
-            .line_separator("\n")
-            .indent_string("\t")
-            .perform_indent(true)
-            .write_document_declaration(false)
-            .normalize_empty_elements(true)
-            .cdata_to_characters(true)
-            .keep_element_names_stack(false)
-            .autopad_comments(true);
+        let opts = XmlWriteOptions::default();
+        XmlWriter::new_with_options(writer, &opts)
+    }
+
+    pub fn new_with_options(writer: W, opts: &XmlWriteOptions) -> XmlWriter<W> {
+        let xml_writer = if opts.indent_amount == 0 {
+            EventWriter::new(writer)
+        } else {
+            EventWriter::new_with_indent(writer, opts.indent_char, opts.indent_amount)
+        };
 
         XmlWriter {
-            xml_writer: EventWriter::new_with_config(writer, config),
+            xml_writer,
+            write_root_element: opts.root_element,
+            started_plist: false,
             stack: Vec::new(),
             expecting_key: false,
-            written_prologue: false,
-            empty_namespace: Namespace::empty(),
+            pending_collection: None,
         }
     }
 
+    #[cfg(feature = "enable_unstable_features_that_may_break_with_minor_version_bumps")]
+    pub fn into_inner(self) -> W {
+        self.xml_writer.into_inner()
+    }
+
     fn write_element_and_value(&mut self, name: &str, value: &str) -> Result<(), Error> {
         self.start_element(name)?;
         self.write_value(value)?;
@@ -63,60 +74,52 @@ impl<W: Write> XmlWriter<W> {
 
     fn start_element(&mut self, name: &str) -> Result<(), Error> {
         self.xml_writer
-            .write(XmlEvent::StartElement {
-                name: Name::local(name),
-                attributes: Cow::Borrowed(&[]),
-                namespace: Cow::Borrowed(&self.empty_namespace),
-            })
-            .map_err(from_xml_error)?;
+            .write_event(XmlEvent::Start(BytesStart::new(name)))?;
         Ok(())
     }
 
     fn end_element(&mut self, name: &str) -> Result<(), Error> {
         self.xml_writer
-            .write(XmlEvent::EndElement {
-                name: Some(Name::local(name)),
-            })
-            .map_err(from_xml_error)?;
+            .write_event(XmlEvent::End(BytesEnd::new(name)))?;
         Ok(())
     }
 
     fn write_value(&mut self, value: &str) -> Result<(), Error> {
         self.xml_writer
-            .write(XmlEvent::Characters(value))
-            .map_err(from_xml_error)?;
+            .write_event(XmlEvent::Text(BytesText::new(value)))?;
         Ok(())
     }
 
-    pub fn into_inner(self) -> W {
-        self.xml_writer.into_inner()
-    }
-
     fn write_event<F: FnOnce(&mut Self) -> Result<(), Error>>(
         &mut self,
         f: F,
     ) -> Result<(), Error> {
-        if !self.written_prologue {
-            self.xml_writer
-                .inner_mut()
-                .write_all(XML_PROLOGUE.as_bytes())
-                .map_err(error::from_io_without_position)?;
+        if !self.started_plist {
+            if self.write_root_element {
+                self.xml_writer
+                    .get_mut()
+                    .write_all(XML_PROLOGUE)
+                    .map_err(error::from_io_without_position)?;
+            }
 
-            self.written_prologue = true;
+            self.started_plist = true;
         }
 
         f(self)?;
 
         // If there are no more open tags then write the </plist> element
         if self.stack.is_empty() {
-            // We didn't tell the xml_writer about the <plist> tag so we'll skip telling it
-            // about the </plist> tag as well.
+            if self.write_root_element {
+                // We didn't tell the xml_writer about the <plist> tag so we'll skip telling it
+                // about the </plist> tag as well.
+                self.xml_writer
+                    .get_mut()
+                    .write_all(b"\n</plist>")
+                    .map_err(error::from_io_without_position)?;
+            }
+
             self.xml_writer
-                .inner_mut()
-                .write_all(b"\n</plist>")
-                .map_err(error::from_io_without_position)?;
-            self.xml_writer
-                .inner_mut()
+                .get_mut()
                 .flush()
                 .map_err(error::from_io_without_position)?;
         }
@@ -129,6 +132,7 @@ impl<W: Write> XmlWriter<W> {
         event_kind: EventKind,
         f: F,
     ) -> Result<(), Error> {
+        self.handle_pending_collection()?;
         self.write_event(|this| {
             if this.expecting_key {
                 return Err(ErrorKind::UnexpectedEventType {
@@ -142,27 +146,61 @@ impl<W: Write> XmlWriter<W> {
             Ok(())
         })
     }
+
+    fn handle_pending_collection(&mut self) -> Result<(), Error> {
+        if let Some(PendingCollection::Array) = self.pending_collection {
+            self.pending_collection = None;
+
+            self.write_value_event(EventKind::StartArray, |this| {
+                this.start_element("array")?;
+                this.stack.push(Element::Array);
+                Ok(())
+            })
+        } else if let Some(PendingCollection::Dictionary) = self.pending_collection {
+            self.pending_collection = None;
+
+            self.write_value_event(EventKind::StartDictionary, |this| {
+                this.start_element("dict")?;
+                this.stack.push(Element::Dictionary);
+                this.expecting_key = true;
+                Ok(())
+            })
+        } else {
+            Ok(())
+        }
+    }
 }
 
 impl<W: Write> Writer for XmlWriter<W> {
     fn write_start_array(&mut self, _len: Option<u64>) -> Result<(), Error> {
-        self.write_value_event(EventKind::StartArray, |this| {
-            this.start_element("array")?;
-            this.stack.push(Element::Array);
-            Ok(())
-        })
+        self.handle_pending_collection()?;
+        self.pending_collection = Some(PendingCollection::Array);
+        Ok(())
     }
 
     fn write_start_dictionary(&mut self, _len: Option<u64>) -> Result<(), Error> {
-        self.write_value_event(EventKind::StartDictionary, |this| {
-            this.start_element("dict")?;
-            this.stack.push(Element::Dictionary);
-            Ok(())
-        })
+        self.handle_pending_collection()?;
+        self.pending_collection = Some(PendingCollection::Dictionary);
+        Ok(())
     }
 
     fn write_end_collection(&mut self) -> Result<(), Error> {
         self.write_event(|this| {
+            match this.pending_collection.take() {
+                Some(PendingCollection::Array) => {
+                    this.xml_writer
+                        .write_event(XmlEvent::Empty(BytesStart::new("array")))?;
+                    this.expecting_key = this.stack.last() == Some(&Element::Dictionary);
+                    return Ok(());
+                }
+                Some(PendingCollection::Dictionary) => {
+                    this.xml_writer
+                        .write_event(XmlEvent::Empty(BytesStart::new("dict")))?;
+                    this.expecting_key = this.stack.last() == Some(&Element::Dictionary);
+                    return Ok(());
+                }
+                _ => {}
+            };
             match (this.stack.pop(), this.expecting_key) {
                 (Some(Element::Dictionary), true) => {
                     this.end_element("dict")?;
@@ -185,13 +223,14 @@ impl<W: Write> Writer for XmlWriter<W> {
 
     fn write_boolean(&mut self, value: bool) -> Result<(), Error> {
         self.write_value_event(EventKind::Boolean, |this| {
-            let value_str = if value { "true" } else { "false" };
-            this.start_element(value_str)?;
-            this.end_element(value_str)
+            let value = if value { "true" } else { "false" };
+            Ok(this
+                .xml_writer
+                .write_event(XmlEvent::Empty(BytesStart::new(value)))?)
         })
     }
 
-    fn write_data(&mut self, value: &[u8]) -> Result<(), Error> {
+    fn write_data(&mut self, value: Cow<[u8]>) -> Result<(), Error> {
         self.write_value_event(EventKind::Data, |this| {
             let base64_data = base64_encode_plist(&value, this.stack.len());
             this.write_element_and_value("data", &base64_data)
@@ -200,7 +239,7 @@ impl<W: Write> Writer for XmlWriter<W> {
 
     fn write_date(&mut self, value: Date) -> Result<(), Error> {
         self.write_value_event(EventKind::Date, |this| {
-            this.write_element_and_value("date", &value.to_rfc3339())
+            this.write_element_and_value("date", &value.to_xml_format())
         })
     }
 
@@ -216,13 +255,14 @@ impl<W: Write> Writer for XmlWriter<W> {
         })
     }
 
-    fn write_string(&mut self, value: &str) -> Result<(), Error> {
+    fn write_string(&mut self, value: Cow<str>) -> Result<(), Error> {
+        self.handle_pending_collection()?;
         self.write_event(|this| {
             if this.expecting_key {
-                this.write_element_and_value("key", &*value)?;
+                this.write_element_and_value("key", &value)?;
                 this.expecting_key = false;
             } else {
-                this.write_element_and_value("string", &*value)?;
+                this.write_element_and_value("string", &value)?;
                 this.expecting_key = this.stack.last() == Some(&Element::Dictionary);
             }
             Ok(())
@@ -234,17 +274,20 @@ impl<W: Write> Writer for XmlWriter<W> {
     }
 }
 
-pub(crate) fn from_xml_error(err: XmlWriterError) -> Error {
-    match err {
-        XmlWriterError::Io(err) => ErrorKind::Io(err).without_position(),
-        XmlWriterError::DocumentStartAlreadyEmitted
-        | XmlWriterError::LastElementNameNotAvailable
-        | XmlWriterError::EndElementNameIsNotEqualToLastStartElementName
-        | XmlWriterError::EndElementNameIsNotSpecified => unreachable!(),
+impl From<XmlWriterError> for Error {
+    fn from(err: XmlWriterError) -> Self {
+        match err {
+            XmlWriterError::Io(err) => match std::sync::Arc::try_unwrap(err) {
+                Ok(err) => ErrorKind::Io(err),
+                Err(err) => ErrorKind::Io(std::io::Error::from(err.kind())),
+            }
+            .without_position(),
+            _ => unreachable!(),
+        }
     }
 }
 
-fn base64_encode_plist(data: &[u8], indent: usize) -> String {
+pub(crate) fn base64_encode_plist(data: &[u8], indent: usize) -> String {
     // XML plist data elements are always formatted by apple tools as
     // <data>
     // AAAA..AA (68 characters per line)
@@ -269,8 +312,9 @@ fn base64_encode_plist(data: &[u8], indent: usize) -> String {
     output[..line_ending.len()].copy_from_slice(&line_ending);
 
     // Encode `data` as a base 64 string
-    let base64_string_len =
-        base64::encode_config_slice(data, base64::STANDARD, &mut output[line_ending.len()..]);
+    let base64_string_len = base64_standard
+        .encode_slice(data, &mut output[line_ending.len()..])
+        .expect("encoding slice fits base64 buffer");
 
     // Line wrap the base 64 encoded string
     let line_wrap_len = line_wrap::line_wrap(
@@ -291,7 +335,6 @@ fn base64_encode_plist(data: &[u8], indent: usize) -> String {
 
 #[cfg(test)]
 mod tests {
-    use humantime::parse_rfc3339_weak;
     use std::io::Cursor;
 
     use super::*;
@@ -299,29 +342,29 @@ mod tests {
 
     #[test]
     fn streaming_parser() {
-        let plist = &[
+        let plist = [
             Event::StartDictionary(None),
-            Event::String("Author".to_owned()),
-            Event::String("William Shakespeare".to_owned()),
-            Event::String("Lines".to_owned()),
+            Event::String("Author".into()),
+            Event::String("William Shakespeare".into()),
+            Event::String("Lines".into()),
             Event::StartArray(None),
-            Event::String("It is a tale told by an idiot,".to_owned()),
-            Event::String("Full of sound and fury, signifying nothing.".to_owned()),
-            Event::Data((0..128).collect::<Vec<_>>()),
+            Event::String("It is a tale told by an idiot,".into()),
+            Event::String("Full of sound and fury, signifying nothing.".into()),
+            Event::Data((0..128).collect::<Vec<_>>().into()),
             Event::EndCollection,
-            Event::String("Death".to_owned()),
+            Event::String("Death".into()),
             Event::Integer(1564.into()),
-            Event::String("Height".to_owned()),
+            Event::String("Height".into()),
             Event::Real(1.60),
-            Event::String("Data".to_owned()),
-            Event::Data(vec![0, 0, 0, 190, 0, 0, 0, 3, 0, 0, 0, 30, 0, 0, 0]),
-            Event::String("Birthdate".to_owned()),
-            Event::Date(parse_rfc3339_weak("1981-05-16 11:32:06").unwrap().into()),
-            Event::String("Comment".to_owned()),
-            Event::String("2 < 3".to_owned()), // make sure characters are escaped
-            Event::String("BiggestNumber".to_owned()),
+            Event::String("Data".into()),
+            Event::Data(vec![0, 0, 0, 190, 0, 0, 0, 3, 0, 0, 0, 30, 0, 0, 0].into()),
+            Event::String("Birthdate".into()),
+            Event::Date(super::Date::from_xml_format("1981-05-16T11:32:06Z").unwrap()),
+            Event::String("Comment".into()),
+            Event::String("2 < 3".into()), // make sure characters are escaped
+            Event::String("BiggestNumber".into()),
             Event::Integer(18446744073709551615u64.into()),
-            Event::String("SmallestNumber".to_owned()),
+            Event::String("SmallestNumber".into()),
             Event::Integer((-9223372036854775808i64).into()),
             Event::String("IsTrue".into()),
             Event::Boolean(true),
@@ -330,17 +373,7 @@ mod tests {
             Event::EndCollection,
         ];
 
-        let mut cursor = Cursor::new(Vec::new());
-
-        {
-            let mut plist_w = XmlWriter::new(&mut cursor);
-
-            for item in plist {
-                plist_w.write(item).unwrap();
-            }
-        }
-
-        let comparison = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>
+        let expected = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>
 <!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">
 <plist version=\"1.0\">
 <dict>
@@ -373,14 +406,68 @@ mod tests {
 \t<key>SmallestNumber</key>
 \t<integer>-9223372036854775808</integer>
 \t<key>IsTrue</key>
-\t<true />
+\t<true/>
 \t<key>IsNotFalse</key>
-\t<false />
+\t<false/>
 </dict>
 </plist>";
 
-        let s = String::from_utf8(cursor.into_inner()).unwrap();
+        let actual = events_to_xml(plist, XmlWriteOptions::default());
 
-        assert_eq!(s, comparison);
+        assert_eq!(actual, expected);
+    }
+
+    #[test]
+    fn custom_indent_string() {
+        let plist = [
+            Event::StartArray(None),
+            Event::String("It is a tale told by an idiot,".into()),
+            Event::String("Full of sound and fury, signifying nothing.".into()),
+            Event::EndCollection,
+        ];
+
+        let expected = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>
+<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//EN\" \"http://www.apple.com/DTDs/PropertyList-1.0.dtd\">
+<plist version=\"1.0\">
+<array>
+...<string>It is a tale told by an idiot,</string>
+...<string>Full of sound and fury, signifying nothing.</string>
+</array>
+</plist>";
+
+        let actual = events_to_xml(plist, XmlWriteOptions::default().indent(b'.', 3));
+
+        assert_eq!(actual, expected);
+    }
+
+    #[test]
+    fn no_root() {
+        let plist = [
+            Event::StartArray(None),
+            Event::String("It is a tale told by an idiot,".into()),
+            Event::String("Full of sound and fury, signifying nothing.".into()),
+            Event::EndCollection,
+        ];
+
+        let expected = "<array>
+\t<string>It is a tale told by an idiot,</string>
+\t<string>Full of sound and fury, signifying nothing.</string>
+</array>";
+
+        let actual = events_to_xml(plist, XmlWriteOptions::default().root_element(false));
+
+        assert_eq!(actual, expected);
+    }
+
+    fn events_to_xml<'event>(
+        events: impl IntoIterator<Item = Event<'event>>,
+        options: XmlWriteOptions,
+    ) -> String {
+        let mut cursor = Cursor::new(Vec::new());
+        let mut writer = XmlWriter::new_with_options(&mut cursor, &options);
+        for event in events {
+            writer.write(event).unwrap();
+        }
+        String::from_utf8(cursor.into_inner()).unwrap()
     }
 }
